import logging
import re
from typing import Iterable, List, Mapping, Optional, Tuple, Type, Union, cast

from sciencebeam_parser.document.semantic_document import (
    SemanticAuthor,
    SemanticContentFactoryProtocol,
    SemanticContentWrapper,
    SemanticMarker,
    SemanticMiddleName,
    SemanticMixedContentWrapper,
    SemanticNamePart,
    SemanticNameSuffix,
    SemanticNameTitle,
    SemanticNote,
    SemanticGivenName,
    SemanticSurname,
    T_SemanticName
)
from sciencebeam_parser.document.layout_document import LayoutBlock, LayoutDocument, LayoutToken
from sciencebeam_parser.models.extract import SimpleModelSemanticExtractor


LOGGER = logging.getLogger(__name__)


SPLIT_ON_SECOND_ENTIY_NAME = {'<title>', '<forename>', '<surname>'}


SIMPLE_SEMANTIC_CONTENT_CLASS_BY_TAG: Mapping[str, SemanticContentFactoryProtocol] = {
    '<title>': SemanticNameTitle,
    '<forename>': SemanticGivenName,
    '<middlename>': SemanticMiddleName,
    '<surname>': SemanticSurname,
    '<suffix>': SemanticNameSuffix
}


def tokenize_individual_characters(text: str) -> List[str]:
    return list(text)


def convert_two_letter_uppercase_given_name_to_given_middle_name(
    name: T_SemanticName
):
    given_names = list(name.iter_by_type(SemanticGivenName))
    middle_names = list(name.iter_by_type(SemanticMiddleName))
    if middle_names:
        LOGGER.debug('already has a middle name: %r', middle_names)
        return
    if len(given_names) != 1:
        LOGGER.debug('no or too many given names: %r', given_names)
        return
    given_name_text = given_names[0].get_text()
    if len(given_name_text) != 2 or not given_name_text.isupper():
        LOGGER.debug('not two uppercase characters: %r', given_name_text)
        return
    layout_document = LayoutDocument.for_blocks(list(given_names[0].iter_blocks()))
    retokenized_layout_document = layout_document.retokenize(
        tokenize_fn=tokenize_individual_characters
    )
    LOGGER.debug('retokenized_layout_document: %r', retokenized_layout_document)
    split_name_parts = [
        (
            SemanticGivenName(layout_block=LayoutBlock.for_tokens([token])) if index == 0
            else SemanticMiddleName(layout_block=LayoutBlock.for_tokens([token]))
        )
        for index, token in enumerate(retokenized_layout_document.iter_all_tokens())
    ]
    LOGGER.debug('split_name_parts: %r', split_name_parts)
    name.flat_map_inplace_by_type(
        SemanticGivenName,
        lambda _: split_name_parts
    )


def convert_name_parts_to_title_case(name: T_SemanticName):
    for semantic_content in name:
        if not isinstance(semantic_content, SemanticNamePart):
            continue
        semantic_content.value = semantic_content.get_text().title()


# based on:
# https://github.com/kermitt2/grobid/blob/0.6.2/grobid-core/src/main/java/org/grobid/core/data/Person.java#L375-L391
# and:
# https://github.com/kermitt2/grobid/blob/0.6.2/grobid-core/src/main/java/org/grobid/core/data/Person.java#L756-L775
def normalize_name_parts(name: T_SemanticName):
    if not list(name.iter_by_type(SemanticSurname)):
        return SemanticNote(
            layout_block=LayoutBlock.merge_blocks(name.iter_blocks()),
            note_type='invalid_author_name'
        )
    convert_two_letter_uppercase_given_name_to_given_middle_name(name)
    convert_name_parts_to_title_case(name)
    return name


def iter_semantic_markers_for_layout_block(
    layout_block: LayoutBlock
) -> Iterable[Union[SemanticMarker, SemanticContentWrapper]]:
    for text in re.split(r'(\D)', layout_block.text):
        if not text:
            continue
        local_block = LayoutBlock.for_tokens([
            LayoutToken(text, whitespace='')
        ])
        if text == ',' or text.isspace():
            yield SemanticNote(
                layout_block=local_block,
                note_type='marker_delimiter'
            )
            continue
        yield SemanticMarker(layout_block=local_block)


def append_semantic_markers_for_layout_block(
    parent_semantic_content: SemanticMixedContentWrapper,
    layout_block: LayoutBlock
) -> None:
    semantic_markers = list(iter_semantic_markers_for_layout_block(layout_block))
    for semantic_marker in semantic_markers:
        parent_semantic_content.add_content(semantic_marker)


class NameSemanticExtractor(SimpleModelSemanticExtractor):
    def __init__(self):
        super().__init__(semantic_content_class_by_tag=SIMPLE_SEMANTIC_CONTENT_CLASS_BY_TAG)

    def iter_semantic_content_for_entity_blocks(  # type: ignore  # pylint: disable=arguments-differ
        self,
        entity_tokens: Iterable[Tuple[str, LayoutBlock]],
        name_type: Optional[Type[T_SemanticName]] = None,
        **kwargs
    ) -> Iterable[T_SemanticName]:
        _name_type: Type[T_SemanticName] = cast(
            Type[T_SemanticName],
            name_type if name_type is not None else SemanticAuthor
        )
        entity_tokens = list(entity_tokens)
        LOGGER.debug('entity_tokens: %s', entity_tokens)
        semantic_name: Optional[T_SemanticName] = None
        seen_entity_tokens: List[Tuple[str, LayoutBlock]] = []
        seen_name_labels: List[str] = []
        has_tail_marker: bool = False
        for name, layout_block in entity_tokens:
            seen_entity_tokens.append((name, layout_block,))
            if name == '<marker>':
                if not semantic_name:
                    LOGGER.debug('new semantic_name with marker in the beginning')
                    semantic_name = _name_type()
                    append_semantic_markers_for_layout_block(semantic_name, layout_block)
                    continue
                if len(seen_entity_tokens) >= 2 and seen_name_labels and not has_tail_marker:
                    previous_layout_block = seen_entity_tokens[-2][1]
                    if previous_layout_block.text.strip().endswith(','):
                        LOGGER.debug(
                            'new semantic_name marker after comma, seen_name_labels=%s',
                            seen_name_labels
                        )
                        yield normalize_name_parts(semantic_name)
                        seen_name_labels = []
                        semantic_name = _name_type()
                        append_semantic_markers_for_layout_block(semantic_name, layout_block)
                        continue
                append_semantic_markers_for_layout_block(semantic_name, layout_block)
                has_tail_marker = True
                continue
            if semantic_name and name in SPLIT_ON_SECOND_ENTIY_NAME and name in seen_name_labels:
                LOGGER.debug(
                    'starting new semantic_name after having seen name part again, name=%r',
                    name
                )
                yield normalize_name_parts(semantic_name)
                seen_name_labels = []
                has_tail_marker = False
                semantic_name = None
            semantic_content = self.get_semantic_content_for_entity_name(
                name, layout_block
            )
            if not isinstance(semantic_content, SemanticNote):
                if has_tail_marker and semantic_name:
                    LOGGER.debug('starting new semantic_name after tail markers, name=%r', name)
                    yield normalize_name_parts(semantic_name)
                    seen_name_labels = []
                    has_tail_marker = False
                    semantic_name = None
                seen_name_labels.append(name)
            if not semantic_name:
                semantic_name = _name_type()
            semantic_name.add_content(semantic_content)
        if semantic_name:
            yield normalize_name_parts(semantic_name)



import logging

from sciencebeam_parser.models.model import Model
from sciencebeam_parser.models.data import (
    DocumentFeaturesContext
)
from sciencebeam_parser.models.affiliation_address.data import AffiliationAddressDataGenerator
from sciencebeam_parser.models.affiliation_address.extract import (
    AffiliationAddressSemanticExtractor
)
from sciencebeam_parser.models.affiliation_address.training_data import (
    AffiliationAddressTeiTrainingDataGenerator,
    AffiliationAddressTrainingTeiParser
)


LOGGER = logging.getLogger(__name__)


class AffiliationAddressModel(Model):
    def get_data_generator(
        self,
        document_features_context: DocumentFeaturesContext
    ) -> AffiliationAddressDataGenerator:
        return AffiliationAddressDataGenerator(
            document_features_context=document_features_context
        )

    def get_semantic_extractor(self) -> AffiliationAddressSemanticExtractor:
        return AffiliationAddressSemanticExtractor()

    def get_tei_training_data_generator(self) -> AffiliationAddressTeiTrainingDataGenerator:
        return AffiliationAddressTeiTrainingDataGenerator()

    def get_training_tei_parser(self) -> AffiliationAddressTrainingTeiParser:
        return AffiliationAddressTrainingTeiParser()



from setuptools import find_packages, setup

import sciencebeam_parser
from sciencebeam_parser.utils.dist.utils import (
    get_markdown_with_resolved_links,
    get_github_content_link_prefix
)


PROJECT_URL = 'https://github.com/elifesciences/sciencebeam-parser'
VERSION = sciencebeam_parser.__version__


with open('requirements.txt', 'r', encoding='utf-8') as f:
    REQUIRED_PACKAGES = f.readlines()


with open('doc/python_library.md', 'r', encoding='utf-8') as f:
    LONG_DESCRIPTION = get_markdown_with_resolved_links(
        '\n'.join([
            line.rstrip()
            for line in f
            if not line.startswith('[![')
        ]),
        source_base_path='doc',
        link_prefix=get_github_content_link_prefix(
            PROJECT_URL,
            version=VERSION
        )
    )


packages = find_packages(exclude=["tests", "tests.*"])

setup(
    name="sciencebeam_parser",
    version=VERSION,
    author="Daniel Ecer",
    url=PROJECT_URL,
    install_requires=REQUIRED_PACKAGES,
    packages=packages,
    include_package_data=True,
    description='ScienceBeam Parser, parse scientific documents.',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)



import dataclasses
import logging
import itertools
import operator
from dataclasses import dataclass, field
from functools import partial
from typing import Callable, List, Iterable, NamedTuple, Optional, Sequence, Tuple

from sciencebeam_parser.utils.bounding_box import BoundingBox
from sciencebeam_parser.utils.tokenizer import iter_tokenized_tokens, get_tokenized_tokens


LOGGER = logging.getLogger(__name__)


class LayoutFont(NamedTuple):
    font_id: str
    font_family: Optional[str] = None
    font_size: Optional[float] = None
    is_bold: Optional[bool] = None
    is_italics: Optional[bool] = None
    is_subscript: Optional[bool] = None
    is_superscript: Optional[bool] = None


EMPTY_FONT = LayoutFont(font_id='_EMPTY')


class LayoutPageCoordinates(NamedTuple):
    x: float
    y: float
    width: float
    height: float
    page_number: int = 0

    @staticmethod
    def from_bounding_box(
        bounding_box: BoundingBox,
        page_number: int = 0
    ) -> 'LayoutPageCoordinates':
        return LayoutPageCoordinates(
            x=bounding_box.x,
            y=bounding_box.y,
            width=bounding_box.width,
            height=bounding_box.height,
            page_number=page_number
        )

    @property
    def bounding_box(self) -> BoundingBox:
        return BoundingBox(x=self.x, y=self.y, width=self.width, height=self.height)

    def __bool__(self) -> bool:
        return not self.is_empty()

    def is_empty(self) -> bool:
        return self.width == 0 or self.height == 0

    def move_by(self, dx: float = 0, dy: float = 0) -> 'LayoutPageCoordinates':
        return LayoutPageCoordinates(
            x=self.x + dx, y=self.y + dy, width=self.width, height=self.height,
            page_number=self.page_number
        )

    def get_merged_with(
        self,
        other: 'LayoutPageCoordinates'
    ) -> 'LayoutPageCoordinates':
        assert self.page_number == other.page_number, \
            'cannot merge coordinates on different pages'
        x = min(self.x, other.x)
        y = min(self.y, other.y)
        width = max(self.x + self.width, other.x + other.width) - x
        height = max(self.y + self.height, other.y + other.height) - y
        return LayoutPageCoordinates(
            x=x, y=y, width=width, height=height, page_number=self.page_number
        )


def get_merged_coordinates_list(
    coordinates_list: Iterable[LayoutPageCoordinates]
) -> List[LayoutPageCoordinates]:
    result: List[LayoutPageCoordinates] = []
    pending_coordinates: Optional[LayoutPageCoordinates] = None
    for coordinates in coordinates_list:
        if not pending_coordinates:
            pending_coordinates = coordinates
            continue
        if coordinates.page_number != pending_coordinates.page_number:
            result.append(pending_coordinates)
            pending_coordinates = coordinates
            continue
        pending_coordinates = pending_coordinates.get_merged_with(
            coordinates
        )
    if pending_coordinates:
        result.append(pending_coordinates)
    return result


class LayoutPageMeta(NamedTuple):
    page_number: int = 0
    coordinates: Optional[LayoutPageCoordinates] = None

    @staticmethod
    def for_coordinates(coordinates: LayoutPageCoordinates) -> 'LayoutPageMeta':
        return LayoutPageMeta(page_number=coordinates.page_number, coordinates=coordinates)


DEFAULT_LAYOUT_PAGE_META = LayoutPageMeta()


class LayoutLineMeta(NamedTuple):
    line_id: int = -1
    page_meta: LayoutPageMeta = DEFAULT_LAYOUT_PAGE_META


DEFAULT_LAYOUT_LINE_META = LayoutLineMeta()


class LayoutToken(NamedTuple):
    text: str
    font: LayoutFont = EMPTY_FONT
    whitespace: str = ' '
    coordinates: Optional[LayoutPageCoordinates] = None
    line_meta: LayoutLineMeta = DEFAULT_LAYOUT_LINE_META


T_FlatMapLayoutTokensFn = Callable[[LayoutToken], List[LayoutToken]]


def default_get_tokenized_tokens_keep_whitespace(text: str) -> List[str]:
    return get_tokenized_tokens(text, keep_whitespace=True)


def get_relative_coordinates(
    coordinates: Optional[LayoutPageCoordinates],
    text: str,
    text_character_offset: int,
    total_text_length: int
) -> Optional[LayoutPageCoordinates]:
    if not coordinates:
        return None
    return LayoutPageCoordinates(
        page_number=coordinates.page_number,
        x=(
            coordinates.x
            + coordinates.width * text_character_offset / total_text_length
        ),
        y=coordinates.y,
        width=(
            coordinates.width
            * len(text) / total_text_length
        ),
        height=coordinates.height
    )


def retokenize_layout_token(
    layout_token: LayoutToken,
    tokenize_fn: Optional[Callable[[str], List[str]]] = None
) -> List[LayoutToken]:
    if not layout_token.text.strip():
        return []
    if tokenize_fn is None:
        tokenize_fn = default_get_tokenized_tokens_keep_whitespace
    token_texts = tokenize_fn(layout_token.text)
    if token_texts == [layout_token.text]:
        return [layout_token]
    total_text_length = sum(len(token_text) for token_text in token_texts)
    texts_with_whitespace: List[Tuple[str, str, int]] = []
    pending_token_text = ''
    pending_whitespace = ''
    text_character_offset = 0
    pending_text_character_offset = 0
    for token_text in token_texts:
        if not token_text.strip():
            pending_whitespace += token_text
            text_character_offset += len(token_text)
            continue
        if pending_token_text:
            texts_with_whitespace.append((
                pending_token_text,
                pending_whitespace,
                pending_text_character_offset
            ))
        pending_token_text = token_text
        pending_whitespace = ''
        pending_text_character_offset = text_character_offset
        text_character_offset += len(token_text)
    pending_whitespace += layout_token.whitespace
    if pending_token_text:
        texts_with_whitespace.append((
            pending_token_text,
            pending_whitespace,
            pending_text_character_offset
        ))
    return [
        LayoutToken(
            text=token_text,
            font=layout_token.font,
            whitespace=whitespace,
            coordinates=get_relative_coordinates(
                layout_token.coordinates,
                pending_token_text,
                text_character_offset,
                total_text_length
            ),
            line_meta=layout_token.line_meta
        )
        for token_text, whitespace, text_character_offset in texts_with_whitespace
    ]


def iter_layout_tokens_for_text(
    text: str,
    tail_whitespace: str = ' ',
    **kwargs
) -> Iterable[LayoutToken]:
    pending_text = ''
    pending_whitespace = ' '
    for token_text in iter_tokenized_tokens(text, keep_whitespace=True):
        if not token_text.strip():
            pending_whitespace += token_text
            continue
        if pending_text:
            yield LayoutToken(pending_text, whitespace=pending_whitespace, **kwargs)
        pending_text = token_text
        pending_whitespace = ''
    if pending_text:
        pending_whitespace += tail_whitespace
        yield LayoutToken(pending_text, whitespace=pending_whitespace, **kwargs)


def get_layout_tokens_for_text(*args, **kwargs) -> List[LayoutToken]:
    return list(iter_layout_tokens_for_text(*args, **kwargs))


@dataclass
class LayoutLine:
    tokens: List[LayoutToken]

    @property
    def text(self) -> str:
        return join_layout_tokens(self.tokens)

    @staticmethod
    def for_text(text: str, **kwargs) -> 'LayoutLine':
        return LayoutLine(tokens=get_layout_tokens_for_text(text, **kwargs))

    def flat_map_layout_tokens(self, fn: T_FlatMapLayoutTokensFn) -> 'LayoutLine':
        return LayoutLine(tokens=[
            tokenized_token
            for token in self.tokens
            for tokenized_token in fn(token)
        ])


@dataclass
class LayoutBlock:
    lines: List[LayoutLine]

    def __len__(self):
        return len(self.lines)

    @staticmethod
    def for_tokens(tokens: List[LayoutToken]) -> 'LayoutBlock':
        if not tokens:
            return EMPTY_BLOCK
        lines = [
            LayoutLine(tokens=list(line_tokens))
            for _, line_tokens in itertools.groupby(
                tokens, key=operator.attrgetter('line_meta')
            )
        ]
        return LayoutBlock(lines=lines)

    @staticmethod
    def merge_blocks(blocks: Iterable['LayoutBlock']) -> 'LayoutBlock':
        return LayoutBlock(lines=[
            line
            for block in blocks
            for line in block.lines
        ])

    @staticmethod
    def for_text(text: str, **kwargs) -> 'LayoutBlock':
        return LayoutBlock(lines=[LayoutLine.for_text(text, **kwargs)])

    def iter_all_tokens(self) -> Iterable[LayoutToken]:
        return (
            token
            for line in self.lines
            for token in line.tokens
        )

    def get_merged_coordinates_list(self) -> List[LayoutPageCoordinates]:
        return get_merged_coordinates_list([
            token.coordinates
            for token in self.iter_all_tokens()
            if token.coordinates
        ])

    def flat_map_layout_tokens(self, fn: T_FlatMapLayoutTokensFn) -> 'LayoutBlock':
        return LayoutBlock(lines=[
            line.flat_map_layout_tokens(fn)
            for line in self.lines
        ])

    def remove_empty_lines(self) -> 'LayoutBlock':
        return LayoutBlock(lines=[
            line
            for line in self.lines
            if line.tokens
        ])

    @property
    def text(self) -> str:
        return join_layout_tokens(self.iter_all_tokens())

    @property
    def whitespace(self) -> str:
        if not self.lines or not self.lines[-1].tokens:
            return ''
        return self.lines[-1].tokens[-1].whitespace


EMPTY_BLOCK = LayoutBlock(lines=[])


class LayoutGraphic(NamedTuple):
    local_file_path: Optional[str] = None
    coordinates: Optional[LayoutPageCoordinates] = None
    graphic_type: Optional[str] = None
    related_block: Optional[LayoutBlock] = None
    page_meta: LayoutPageMeta = DEFAULT_LAYOUT_PAGE_META


@dataclass
class LayoutPage:
    blocks: List[LayoutBlock]
    graphics: Sequence[LayoutGraphic] = field(default_factory=list)
    meta: LayoutPageMeta = DEFAULT_LAYOUT_PAGE_META

    def replace(self, **changes) -> 'LayoutPage':
        return dataclasses.replace(self, **changes)

    def iter_all_tokens(self) -> Iterable[LayoutToken]:
        return (
            token
            for block in self.blocks
            for token in block.iter_all_tokens()
        )

    def flat_map_layout_tokens(self, fn: T_FlatMapLayoutTokensFn) -> 'LayoutPage':
        return LayoutPage(
            blocks=[
                block.flat_map_layout_tokens(fn)
                for block in self.blocks
            ],
            graphics=self.graphics,
            meta=self.meta
        )

    def remove_empty_blocks(self) -> 'LayoutPage':
        blocks: List[LayoutBlock] = [
            block.remove_empty_lines()
            for block in self.blocks
        ]
        return LayoutPage(
            blocks=[
                block
                for block in blocks
                if block.lines
            ],
            graphics=self.graphics,
            meta=self.meta
        )


@dataclass
class LayoutDocument:
    pages: List[LayoutPage]

    def __len__(self):
        return len(self.pages)

    @staticmethod
    def for_blocks(blocks: List[LayoutBlock]) -> 'LayoutDocument':
        return LayoutDocument(pages=[LayoutPage(
            blocks=blocks, graphics=[]
        )])

    def replace(self, **changes) -> 'LayoutDocument':
        return dataclasses.replace(self, **changes)

    def iter_all_blocks(self) -> Iterable[LayoutBlock]:
        return (
            block
            for page in self.pages
            for block in page.blocks
        )

    def iter_all_lines(self) -> Iterable[LayoutLine]:
        return (
            line
            for block in self.iter_all_blocks()
            for line in block.lines
        )

    def iter_all_tokens(self) -> Iterable[LayoutToken]:
        return (
            token
            for block in self.iter_all_blocks()
            for token in block.iter_all_tokens()
        )

    def iter_all_graphics(self) -> Iterable[LayoutGraphic]:
        return (
            graphic
            for page in self.pages
            for graphic in page.graphics
        )

    def flat_map_layout_tokens(
        self, fn: T_FlatMapLayoutTokensFn, **kwargs
    ) -> 'LayoutDocument':
        if kwargs:
            fn = partial(fn, **kwargs)
        return LayoutDocument(pages=[
            page.flat_map_layout_tokens(fn)
            for page in self.pages
        ])

    def retokenize(self, **kwargs) -> 'LayoutDocument':
        return self.flat_map_layout_tokens(retokenize_layout_token, **kwargs)

    def remove_empty_blocks(self, preserve_empty_pages: bool = False) -> 'LayoutDocument':
        pages: List[LayoutPage] = [
            page.remove_empty_blocks()
            for page in self.pages
        ]
        return LayoutDocument(pages=[
            page
            for page in pages
            if page.blocks or preserve_empty_pages
        ])


class LayoutTokenIndexRange(NamedTuple):
    layout_token: LayoutToken
    start: int
    end: int


class LayoutTokensText:
    def __init__(self, layout_block: LayoutBlock) -> None:
        self.layout_block = layout_block
        text_fragments = []
        pending_whitespace = ''
        text_offset = 0
        token_index_ranges: List[LayoutTokenIndexRange] = []
        for line in layout_block.lines:
            for token in line.tokens:
                if pending_whitespace:
                    text_fragments.append(pending_whitespace)
                    text_offset += len(pending_whitespace)
                    pending_whitespace = ''
                token_text = token.text
                token_index_ranges.append(LayoutTokenIndexRange(
                    layout_token=token,
                    start=text_offset,
                    end=text_offset + len(token_text)
                ))
                text_fragments.append(token_text)
                text_offset += len(token_text)
                pending_whitespace += token.whitespace
        self.token_index_ranges = token_index_ranges
        self.text = ''.join(text_fragments)

    def __str__(self):
        return self.text

    def iter_layout_tokens_between(
        self, start: int, end: int
    ) -> Iterable[LayoutToken]:
        for token_index_range in self.token_index_ranges:
            if token_index_range.start >= end:
                break
            if token_index_range.end <= start:
                continue
            yield token_index_range.layout_token

    def get_layout_tokens_between(
        self, start: int, end: int
    ) -> List[LayoutToken]:
        return list(self.iter_layout_tokens_between(start, end))


def join_layout_tokens(layout_tokens: Iterable[LayoutToken]) -> str:
    layout_tokens = list(layout_tokens)
    return ''.join([
        (
            token.text + token.whitespace
            if index < len(layout_tokens) - 1
            else token.text
        )
        for index, token in enumerate(layout_tokens)
    ])


def flat_map_layout_document_tokens(
    layout_document: LayoutDocument,
    fn: T_FlatMapLayoutTokensFn,
    **kwargs
) -> LayoutDocument:
    return layout_document.flat_map_layout_tokens(fn, **kwargs)


def retokenize_layout_document(
    layout_document: LayoutDocument,
    **kwargs
) -> LayoutDocument:
    return layout_document.retokenize(**kwargs)


def remove_empty_blocks(
    layout_document: LayoutDocument,
    **kwargs
) -> LayoutDocument:
    return layout_document.remove_empty_blocks(**kwargs)



import logging
from typing import (
    Dict,
    List,
    Optional,
)

from sciencebeam_parser.utils.stop_watch import StopWatchRecorder
from sciencebeam_parser.document.semantic_document import (
    SemanticAbstract,
    SemanticAffiliationAddress,
    SemanticAuthor,
    SemanticDocument,
    SemanticFigure,
    SemanticMarker,
    SemanticNote,
    SemanticReferenceList,
    SemanticSection,
    SemanticSectionTypes,
    SemanticTable,
    SemanticTitle
)
from sciencebeam_parser.document.tei.common import (
    TeiElementBuilder
)
from sciencebeam_parser.document.tei.document import (
    TeiDocument
)
from sciencebeam_parser.document.tei.factories import (
    DEFAULT_TEI_ELEMENT_FACTORY_CONTEXT,
    TeiElementFactoryContext,
    get_tei_note_for_semantic_content
)
from sciencebeam_parser.document.tei.author import (
    get_dummy_tei_author_for_semantic_affiliations_element,
    get_orphan_affiliations,
    get_tei_author_for_semantic_author_element
)
from sciencebeam_parser.document.tei.references import (
    get_tei_raw_reference_list_element
)


LOGGER = logging.getLogger(__name__)


def get_tei_for_semantic_document(  # pylint: disable=too-many-branches, too-many-statements
    semantic_document: SemanticDocument,
    context: Optional[TeiElementFactoryContext] = None
) -> TeiDocument:
    if context is None:
        context = DEFAULT_TEI_ELEMENT_FACTORY_CONTEXT
    stop_watch_recorder = StopWatchRecorder()
    LOGGER.info('generating tei document')
    LOGGER.debug('semantic_document: %s', semantic_document)
    tei_document = TeiDocument()

    stop_watch_recorder.start('front')
    LOGGER.info('generating tei document: front')
    title_block = semantic_document.front.view_by_type(SemanticTitle).merged_block
    if title_block:
        tei_document.set_title_layout_block(title_block)

    abstract_block = semantic_document.front.view_by_type(SemanticAbstract).merged_block
    if abstract_block:
        tei_document.set_abstract_layout_block(abstract_block)

    affiliations_by_marker: Dict[str, List[SemanticAffiliationAddress]] = {}
    for semantic_content in semantic_document.front:
        if isinstance(semantic_content, SemanticAffiliationAddress):
            marker_text = semantic_content.get_text_by_type(SemanticMarker)
            affiliations_by_marker.setdefault(marker_text, []).append(semantic_content)
    LOGGER.debug('affiliations_by_marker: %r', affiliations_by_marker)

    semantic_authors: List[SemanticAuthor] = []
    for semantic_content in semantic_document.front:
        if isinstance(semantic_content, SemanticAuthor):
            semantic_authors.append(semantic_content)
            tei_author_element = get_tei_author_for_semantic_author_element(
                semantic_content,
                context=context,
                affiliations_by_marker=affiliations_by_marker
            )
            tei_document.get_or_create_element_at(
                ['teiHeader', 'fileDesc', 'sourceDesc', 'biblStruct', 'analytic']
            ).append(tei_author_element)
            continue
        if isinstance(semantic_content, SemanticTitle):
            continue
        if isinstance(semantic_content, SemanticAbstract):
            continue
        if isinstance(semantic_content, SemanticAffiliationAddress):
            continue
        tei_document.get_or_create_element_at(
            ['teiHeader']
        ).append(get_tei_note_for_semantic_content(
            semantic_content
        ))
    orphan_affiliations = get_orphan_affiliations(
        affiliations_by_marker=affiliations_by_marker,
        authors=semantic_authors
    )
    if orphan_affiliations:
        dummy_tei_author_element = get_dummy_tei_author_for_semantic_affiliations_element(
            orphan_affiliations,
            context=context
        )
        tei_document.get_or_create_element_at(
            ['teiHeader', 'fileDesc', 'sourceDesc', 'biblStruct', 'analytic']
        ).append(dummy_tei_author_element)

    LOGGER.info('generating tei document: body')
    stop_watch_recorder.start('body')
    for semantic_content in semantic_document.body_section:
        if isinstance(semantic_content, SemanticNote):
            tei_document.get_body().add_note(
                semantic_content.note_type,
                semantic_content.merged_block
            )
            continue
        TeiElementBuilder(tei_document.get_body_element()).extend(
            context.get_tei_child_elements_for_semantic_content(semantic_content)
        )
    for semantic_content in semantic_document.body_section.iter_by_types_recursively(
        (SemanticFigure, SemanticTable,)
    ):
        TeiElementBuilder(tei_document.get_body_element()).extend(
            context.get_tei_child_elements_for_semantic_content(semantic_content)
        )

    LOGGER.info('generating tei document: back section')
    stop_watch_recorder.start('back')
    for semantic_content in semantic_document.back_section:
        if isinstance(semantic_content, SemanticSection):
            if semantic_content.section_type == SemanticSectionTypes.ACKNOWLEDGEMENT:
                _parent = TeiElementBuilder(tei_document.get_back_element())
            else:
                _parent = TeiElementBuilder(tei_document.get_back_annex_element())
            TeiElementBuilder(_parent).extend(
                context.get_tei_child_elements_for_semantic_content(semantic_content)
            )
            continue
        if isinstance(semantic_content, SemanticNote):
            tei_document.get_back_annex().add_note(
                semantic_content.note_type,
                semantic_content.merged_block
            )
            continue
        if isinstance(semantic_content, SemanticReferenceList):
            tei_document.get_references().element.append(
                get_tei_raw_reference_list_element(semantic_content, context=context)
            )
            continue
        TeiElementBuilder(tei_document.get_back_annex_element()).extend(
            context.get_tei_child_elements_for_semantic_content(semantic_content)
        )
    for semantic_figure in semantic_document.back_section.iter_by_types_recursively(
        (SemanticFigure, SemanticTable,)
    ):
        TeiElementBuilder(tei_document.get_back_annex_element()).extend(
            context.get_tei_child_elements_for_semantic_content(semantic_figure)
        )
    stop_watch_recorder.stop()
    LOGGER.info('generating tei document done, took: %s', stop_watch_recorder)
    return tei_document



import logging
from typing import Iterable, List

from lxml.builder import ElementMaker

from sciencebeam_parser.utils.xml_writer import XmlTreeWriter
from sciencebeam_parser.document.layout_document import (
    LayoutLine,
    LayoutToken,
    join_layout_tokens
)
from sciencebeam_parser.models.data import LayoutModelData
from sciencebeam_parser.models.model import (
    get_split_prefix_label
)
from sciencebeam_parser.models.training_data import (
    AbstractTeiTrainingDataGenerator,
    AbstractTrainingTeiParser,
    get_model_data_label
)


LOGGER = logging.getLogger(__name__)


TEI_E = ElementMaker()


# based on:
# https://github.com/kermitt2/grobid/blob/0.7.0/grobid-core/src/main/java/org/grobid/core/engines/Segmentation.java
ROOT_TRAINING_XML_ELEMENT_PATH = ['text']
TRAINING_XML_ELEMENT_PATH_BY_LABEL = {
    '<other>': ROOT_TRAINING_XML_ELEMENT_PATH,
    'O': ROOT_TRAINING_XML_ELEMENT_PATH,
    '<header>': ROOT_TRAINING_XML_ELEMENT_PATH + ['front'],
    '<headnote>': ROOT_TRAINING_XML_ELEMENT_PATH + ['note[@place="headnote"]'],
    '<footnote>': ROOT_TRAINING_XML_ELEMENT_PATH + ['note[@place="footnote"]'],
    '<marginnote>': ROOT_TRAINING_XML_ELEMENT_PATH + ['note[@place="marginnote"]'],
    '<page>': ROOT_TRAINING_XML_ELEMENT_PATH + ['page'],
    '<references>': ROOT_TRAINING_XML_ELEMENT_PATH + ['listBibl'],
    '<body>': ROOT_TRAINING_XML_ELEMENT_PATH + ['body'],
    '<cover>': ROOT_TRAINING_XML_ELEMENT_PATH + ['titlePage'],
    '<toc>': ROOT_TRAINING_XML_ELEMENT_PATH + ['div[@type="toc"]'],
    '<annex>': ROOT_TRAINING_XML_ELEMENT_PATH + ['div[@type="annex"]'],
    '<acknowledgement>': ROOT_TRAINING_XML_ELEMENT_PATH + ['div[@type="acknowledgement"]'],
}


def iter_tokens_from_model_data(model_data: LayoutModelData) -> Iterable[LayoutToken]:
    if model_data.layout_token is not None:
        yield model_data.layout_token
        return
    assert model_data.layout_line is not None
    yield from model_data.layout_line.tokens


def iter_layout_lines_from_layout_tokens(
    layout_tokens: Iterable[LayoutToken]
) -> Iterable[LayoutLine]:
    line_layout_tokens: List[LayoutToken] = []
    for layout_token in layout_tokens:
        if not line_layout_tokens:
            line_layout_tokens.append(layout_token)
            continue
        if (
            layout_token.line_meta.line_id
            == line_layout_tokens[0].line_meta.line_id
        ):
            LOGGER.debug('line id matching: %r - %r', layout_token, line_layout_tokens[0])
            line_layout_tokens.append(layout_token)
            continue
        yield LayoutLine(tokens=line_layout_tokens)
        line_layout_tokens = [layout_token]
    if line_layout_tokens:
        yield LayoutLine(tokens=line_layout_tokens)


class SegmentationTeiTrainingDataGenerator(AbstractTeiTrainingDataGenerator):
    DEFAULT_TEI_FILENAME_SUFFIX = '.segmentation.tei.xml'
    DEFAULT_DATA_FILENAME_SUFFIX = '.segmentation'

    def __init__(self):
        super().__init__(
            root_training_xml_element_path=ROOT_TRAINING_XML_ELEMENT_PATH,
            training_xml_element_path_by_label=TRAINING_XML_ELEMENT_PATH_BY_LABEL,
            element_maker=TEI_E,
            default_tei_filename_suffix=(
                SegmentationTeiTrainingDataGenerator.DEFAULT_TEI_FILENAME_SUFFIX
            ),
            default_data_filename_suffix=(
                SegmentationTeiTrainingDataGenerator.DEFAULT_DATA_FILENAME_SUFFIX
            ),
            default_tei_sub_directory='segmentation/corpus/tei',
            default_data_sub_directory='segmentation/corpus/raw'
        )

    def write_xml_line_for_layout_tokens(
        self,
        xml_writer: XmlTreeWriter,
        layout_tokens: Iterable[LayoutToken]
    ):
        xml_writer.append_text(join_layout_tokens(layout_tokens))
        xml_writer.append(TEI_E('lb'))

    def write_xml_for_model_data_iterable(
        self,
        xml_writer: XmlTreeWriter,
        model_data_iterable: Iterable[LayoutModelData]
    ):
        default_path = xml_writer.current_path
        pending_whitespace = ''
        for model_data in model_data_iterable:
            prefixed_label = get_model_data_label(model_data)
            _prefix, label = get_split_prefix_label(prefixed_label or '')
            xml_element_path = self.get_training_xml_path_for_label(
                label,
                current_path=xml_writer.current_path
            )
            LOGGER.debug('label: %r (%r)', label, xml_element_path)
            if xml_writer.current_path != xml_element_path:
                xml_writer.require_path(default_path)
            xml_writer.append_text(pending_whitespace)
            pending_whitespace = ''
            xml_writer.require_path(xml_element_path)
            for layout_line in iter_layout_lines_from_layout_tokens(
                iter_tokens_from_model_data(model_data)
            ):
                self.write_xml_line_for_layout_tokens(
                    xml_writer,
                    layout_line.tokens
                )
                pending_whitespace = '\n'
        xml_writer.require_path(default_path)
        xml_writer.append_text(pending_whitespace)


class SegmentationTrainingTeiParser(AbstractTrainingTeiParser):
    def __init__(self) -> None:
        super().__init__(
            root_training_xml_element_path=ROOT_TRAINING_XML_ELEMENT_PATH,
            training_xml_element_path_by_label=TRAINING_XML_ELEMENT_PATH_BY_LABEL,
            use_tei_namespace=False,
            line_as_token=True
        )



import logging

from lxml import etree

from sciencebeam_parser.document.tei.common import tei_xpath
from sciencebeam_parser.models.training_data import (
    AbstractTeiTrainingDataGenerator,
    AbstractTrainingTeiParser
)
from sciencebeam_parser.models.citation.extract import (
    get_detected_external_identifier_type_for_text
)
from sciencebeam_parser.utils.xml import get_text_content


LOGGER = logging.getLogger(__name__)


# based on:
# https://github.com/kermitt2/grobid/blob/0.7.0/grobid-core/src/main/java/org/grobid/core/engines/CitationParser.java

ROOT_TRAINING_XML_ELEMENT_PATH = ['text', 'back', 'listBibl', 'bibl']

TRAINING_XML_ELEMENT_PATH_BY_LABEL = {
    '<title>': ROOT_TRAINING_XML_ELEMENT_PATH + ['title[@level="a"]'],
    '<author>': ROOT_TRAINING_XML_ELEMENT_PATH + ['author'],
    '<editor>': ROOT_TRAINING_XML_ELEMENT_PATH + ['editor'],
    '<institution>': ROOT_TRAINING_XML_ELEMENT_PATH + ['orgName'],
    '<collaboration>': ROOT_TRAINING_XML_ELEMENT_PATH + ['orgName[@type="collaboration"]'],
    '<journal>': ROOT_TRAINING_XML_ELEMENT_PATH + ['title[@level="j"]'],
    '<series>': ROOT_TRAINING_XML_ELEMENT_PATH + ['title[@level="s"]'],
    '<booktitle>': ROOT_TRAINING_XML_ELEMENT_PATH + ['title[@level="m"]'],
    '<date>': ROOT_TRAINING_XML_ELEMENT_PATH + ['date'],
    '<volume>': ROOT_TRAINING_XML_ELEMENT_PATH + ['biblScope[@unit="volume"]'],
    '<issue>': ROOT_TRAINING_XML_ELEMENT_PATH + ['biblScope[@unit="issue"]'],
    '<pages>': ROOT_TRAINING_XML_ELEMENT_PATH + ['biblScope[@unit="page"]'],
    '<publisher>': ROOT_TRAINING_XML_ELEMENT_PATH + ['publisher'],
    '<location>': ROOT_TRAINING_XML_ELEMENT_PATH + ['pubPlace'],
    '<tech>': ROOT_TRAINING_XML_ELEMENT_PATH + ['note[@type="report"]'],
    '<web>': ROOT_TRAINING_XML_ELEMENT_PATH + ['ptr[@type="web"]'],
    '<pubnum>': ROOT_TRAINING_XML_ELEMENT_PATH + ['idno'],
    '<note>': ROOT_TRAINING_XML_ELEMENT_PATH + ['note']
}


class CitationTeiTrainingDataGenerator(AbstractTeiTrainingDataGenerator):
    DEFAULT_TEI_FILENAME_SUFFIX = '.references.tei.xml'

    def __init__(self):
        super().__init__(
            root_training_xml_element_path=ROOT_TRAINING_XML_ELEMENT_PATH,
            training_xml_element_path_by_label=TRAINING_XML_ELEMENT_PATH_BY_LABEL,
            use_tei_namespace=True,
            root_tag='TEI',
            default_tei_filename_suffix=(
                CitationTeiTrainingDataGenerator.DEFAULT_TEI_FILENAME_SUFFIX
            ),
            default_data_filename_suffix=None,
            default_tei_sub_directory='citation/corpus'
        )

    def get_post_processed_xml_root(self, xml_root: etree.ElementBase):
        for idno_element in tei_xpath(xml_root, '//tei:idno'):
            external_identifier_type = get_detected_external_identifier_type_for_text(
                get_text_content(idno_element)
            )
            if not external_identifier_type:
                continue
            idno_element.attrib['type'] = external_identifier_type
        return xml_root


class CitationTrainingTeiParser(AbstractTrainingTeiParser):
    def __init__(self) -> None:
        super().__init__(
            root_training_xml_element_path=ROOT_TRAINING_XML_ELEMENT_PATH,
            training_xml_element_path_by_label=(
                TRAINING_XML_ELEMENT_PATH_BY_LABEL
            ),
            use_tei_namespace=True
        )



import logging

from sciencebeam_parser.models.data import (
    DocumentFeaturesContext
)
from sciencebeam_parser.models.model import Model
from sciencebeam_parser.models.reference_segmenter.data import ReferenceSegmenterDataGenerator
from sciencebeam_parser.models.reference_segmenter.extract import (
    ReferenceSegmenterSemanticExtractor
)
from sciencebeam_parser.models.reference_segmenter.training_data import (
    ReferenceSegmenterTeiTrainingDataGenerator,
    ReferenceSegmenterTrainingTeiParser
)


LOGGER = logging.getLogger(__name__)


class ReferenceSegmenterModel(Model):
    def get_data_generator(
        self,
        document_features_context: DocumentFeaturesContext
    ) -> ReferenceSegmenterDataGenerator:
        return ReferenceSegmenterDataGenerator(
            document_features_context=document_features_context
        )

    def get_semantic_extractor(self) -> ReferenceSegmenterSemanticExtractor:
        return ReferenceSegmenterSemanticExtractor()

    def get_tei_training_data_generator(self) -> ReferenceSegmenterTeiTrainingDataGenerator:
        return ReferenceSegmenterTeiTrainingDataGenerator()

    def get_training_tei_parser(self) -> ReferenceSegmenterTrainingTeiParser:
        return ReferenceSegmenterTrainingTeiParser()



from typing import NamedTuple, Set

from sciencebeam_parser.config.config import AppConfig

from sciencebeam_parser.processors.document_page_image import (
    DEFAULT_PDF_RENDER_DPI
)
from sciencebeam_parser.processors.graphic_matching import DEFAULT_MAX_GRAPHIC_DISTANCE


class RequestFieldNames:
    """
    "Abstract" field names that should be independent from the model architecture.
    """
    TITLE = 'title'
    ABSTRACT = 'abstract'
    AUTHORS = 'authors'
    AFFILIATIONS = 'affiliations'
    REFERENCES = 'references'


FRONT_FIELDS = {
    RequestFieldNames.TITLE,
    RequestFieldNames.ABSTRACT,
    RequestFieldNames.AUTHORS,
    RequestFieldNames.AFFILIATIONS
}


class FullTextProcessorConfig(NamedTuple):
    extract_front: bool = True
    extract_authors: bool = True
    extract_affiliations: bool = True
    extract_body_sections: bool = True
    extract_acknowledgements: bool = True
    extract_back_sections: bool = True
    extract_references: bool = True
    extract_citation_fields: bool = True
    extract_citation_authors: bool = True
    extract_citation_editors: bool = False
    extract_figure_fields: bool = True
    extract_table_fields: bool = True
    merge_raw_authors: bool = False
    extract_graphic_bounding_boxes: bool = True
    extract_graphic_assets: bool = False
    use_cv_model: bool = False
    cv_render_dpi: float = DEFAULT_PDF_RENDER_DPI
    use_ocr_model: bool = False
    replace_text_by_cv_graphic: bool = False
    max_graphic_distance: float = DEFAULT_MAX_GRAPHIC_DISTANCE

    @staticmethod
    def from_app_config(app_config: AppConfig) -> 'FullTextProcessorConfig':
        return FullTextProcessorConfig()._replace(
            **app_config.get('processors', {}).get('fulltext', {})
        )

    def get_for_requested_field_names(
        self,
        request_field_names: Set[str]
    ) -> 'FullTextProcessorConfig':
        if not request_field_names:
            return self
        remaining_field_names = request_field_names - FRONT_FIELDS - {RequestFieldNames.REFERENCES}
        if remaining_field_names:
            return self
        extract_front = bool(FRONT_FIELDS & request_field_names)
        extract_authors = RequestFieldNames.AUTHORS in request_field_names
        extract_affiliations = RequestFieldNames.AFFILIATIONS in request_field_names
        extract_references = RequestFieldNames.REFERENCES in request_field_names
        return self._replace(  # pylint: disable=no-member
            extract_front=extract_front,
            extract_authors=extract_authors,
            extract_affiliations=extract_affiliations,
            extract_body_sections=False,
            extract_acknowledgements=False,
            extract_back_sections=False,
            extract_references=extract_references,
            extract_graphic_bounding_boxes=False
        )

    def get_for_header_document(self) -> 'FullTextProcessorConfig':
        return self.get_for_requested_field_names(FRONT_FIELDS)



import logging

from lxml import etree

from sciencebeam_parser.document.semantic_document import (
    SemanticCaption,
    SemanticContentWrapper,
    SemanticFigure,
    SemanticLabel,
    SemanticTable
)
from sciencebeam_parser.document.tei.common import (
    TEI_E
)
from sciencebeam_parser.document.tei.factory import (
    SingleElementTeiElementFactory,
    TeiElementFactoryContext
)


LOGGER = logging.getLogger(__name__)


class FigureTeiElementFactory(SingleElementTeiElementFactory):
    def get_tei_element_for_semantic_content(
        self,
        semantic_content: SemanticContentWrapper,
        context: TeiElementFactoryContext
    ) -> etree.ElementBase:
        LOGGER.debug('semantic_content: %s', semantic_content)
        assert isinstance(semantic_content, SemanticFigure)
        semantic_figure = semantic_content
        children = [context.get_default_attributes_for_semantic_content(semantic_figure)]
        for child_semantic_content in semantic_figure:
            if isinstance(child_semantic_content, SemanticLabel):
                layout_block = child_semantic_content.merged_block
                children.append(TEI_E(
                    'head', *context.iter_layout_block_tei_children(layout_block)
                ))
                children.append(TEI_E(
                    'label', *context.iter_layout_block_tei_children(layout_block)
                ))
                continue
            if isinstance(child_semantic_content, SemanticCaption):
                children.append(TEI_E(
                    'figDesc', *context.iter_layout_block_tei_children(
                        child_semantic_content.merged_block
                    )
                ))
                continue
            children.extend(context.get_tei_child_elements_for_semantic_content(
                child_semantic_content
            ))
        return TEI_E('figure', *children)


class TableTeiElementFactory(SingleElementTeiElementFactory):
    def get_tei_element_for_semantic_content(
        self,
        semantic_content: SemanticContentWrapper,
        context: TeiElementFactoryContext
    ) -> etree.ElementBase:
        LOGGER.debug('semantic_content: %s', semantic_content)
        assert isinstance(semantic_content, SemanticTable)
        semantic_table = semantic_content
        children = [context.get_default_attributes_for_semantic_content(semantic_table)]
        for child_semantic_content in semantic_table:
            if isinstance(child_semantic_content, SemanticLabel):
                layout_block = child_semantic_content.merged_block
                children.append(TEI_E(
                    'head', *context.iter_layout_block_tei_children(layout_block)
                ))
                children.append(TEI_E(
                    'label', *context.iter_layout_block_tei_children(layout_block)
                ))
                continue
            if isinstance(child_semantic_content, SemanticCaption):
                children.append(TEI_E(
                    'figDesc', *context.iter_layout_block_tei_children(
                        child_semantic_content.merged_block
                    )
                ))
                continue
            children.extend(
                context.get_tei_child_elements_for_semantic_content(child_semantic_content)
            )
        return TEI_E('figure', {'type': 'table'}, *children)



import logging
from typing import Any, Mapping, Optional, Union

from lxml import etree


LOGGER = logging.getLogger(__name__)


T_XSLT_Input = Union[etree.ElementBase, etree.ElementTree]


class XsltTransformerWrapper:
    def __init__(
        self,
        xslt_template: str,
        xslt_template_parameters: Optional[Mapping[str, Any]] = None
    ):
        self.xslt_template = xslt_template
        if xslt_template_parameters is None:
            xslt_template_parameters = {}
        self.xslt_template_parameters = xslt_template_parameters
        self.__transformer: Optional[etree.XSLT] = None
        # validate the XSLT stylesheet
        etree.fromstring(self.xslt_template)

    @staticmethod
    def from_template_string(xslt_template: str, **kwargs) -> 'XsltTransformerWrapper':
        return XsltTransformerWrapper(xslt_template, **kwargs)

    @staticmethod
    def from_template_file(xslt_template_file: str, **kwargs) -> 'XsltTransformerWrapper':
        return XsltTransformerWrapper.from_template_string(
            etree.tostring(etree.parse(xslt_template_file)),
            **kwargs
        )

    def _get_transformer(self) -> etree.XSLT:
        if self.__transformer is None:
            # The transform function cannot be pickled and needs to be loaded lazily
            transform = etree.XSLT(
                etree.fromstring(self.xslt_template)
            )
            self.__transformer = transform
        return self.__transformer

    def __call__(
        self,
        xslt_input: T_XSLT_Input,
        xslt_template_parameters: Optional[Mapping[str, Any]] = None
    ):
        xslt_template_parameters = {
            **self.xslt_template_parameters,
            **(xslt_template_parameters or {})
        }
        LOGGER.debug(
            'xslt_input: %r (xslt_template_parameters=%r)',
            xslt_input, xslt_template_parameters
        )
        _xslt_transformer = self._get_transformer()
        return _xslt_transformer(
            xslt_input,
            **{
                key: etree.XSLT.strparam(value)
                for key, value in xslt_template_parameters.items()
            }
        )



import functools

from sciencebeam_parser.ocr_models.ocr_model import (
    LazyOpticalCharacterRecognitionModel,
    OpticalCharacterRecognitionModel
)


class EngineNames:
    TESSEROCR = 'tesserocr'


def get_tesserocr_model(config: dict) -> OpticalCharacterRecognitionModel:
    from sciencebeam_parser.ocr_models.tesserocr_ocr_model import (  # noqa pylint: disable=import-outside-toplevel
        TesserComputerVisionModel
    )
    return TesserComputerVisionModel(config=config)


def get_engine_name_for_config(config: dict) -> str:
    engine_name = config.get('engine')
    if engine_name:
        return engine_name
    return EngineNames.TESSEROCR


def get_ocr_model_for_config(config: dict) -> OpticalCharacterRecognitionModel:
    engine_name = get_engine_name_for_config(config)
    if engine_name == EngineNames.TESSEROCR:
        return get_tesserocr_model(config=config)
    raise RuntimeError('invalid engine name: %r' % engine_name)


def get_lazy_ocr_model_for_config(config: dict) -> OpticalCharacterRecognitionModel:
    return LazyOpticalCharacterRecognitionModel(
        functools.partial(get_ocr_model_for_config, config)
    )



import logging
import os
import copy
from pathlib import Path
from typing import Any, Optional, Union

import yaml


LOGGER = logging.getLogger(__name__)


DEFAULT_DOWNLOAD_DIR = 'data/download'


def parse_env_value(value: str) -> Union[str, int]:
    return yaml.safe_load(value)


class AppConfig:
    def __init__(self, props: dict):
        self.props = props

    def __repr__(self) -> str:
        return '%s(%r)' % (type(self).__name__, self.props)

    @staticmethod
    def load_yaml(file_path: str) -> 'AppConfig':
        return AppConfig(
            yaml.safe_load(Path(file_path).read_text(encoding='utf-8'))
        )

    def apply_environment_variables(self, prefix: str = 'SCIENCEBEAM_PARSER__') -> 'AppConfig':
        updated_props = copy.deepcopy(self.props)
        env_vars = os.environ
        if not env_vars:
            LOGGER.debug('no environment variables')
        LOGGER.debug('processing env vars: %r', env_vars)
        for env_name, env_value in env_vars.items():
            if not env_name.startswith(prefix):
                LOGGER.debug('ignoring: %r', env_name)
                continue
            key_path = env_name[len(prefix):].lower().split('__')
            LOGGER.debug('updating: %r -> %r', env_name, key_path)
            parent_key_path = key_path[:-1]
            leaf_key = key_path[-1]
            parent_props = updated_props
            for parent_key in parent_key_path:
                parent_props = parent_props.setdefault(parent_key, {})
            parent_props[leaf_key] = parse_env_value(env_value)
        return AppConfig(updated_props)

    def get(self, key: str, default_value: Optional[Any] = None):
        return self.props.get(key, default_value)

    def __getitem__(self, key: str):
        return self.props[key]


def get_download_dir(config: Union[dict, AppConfig]) -> str:
    return os.path.expanduser(
        config.get('download_dir', DEFAULT_DOWNLOAD_DIR)
    )



from typing import List, Optional

from lxml import etree

from sciencebeam_parser.document.layout_document import LayoutBlock
from sciencebeam_parser.document.tei.common import (
    TEI_E,
    TeiElementWrapper,
    extend_element,
    get_or_create_element_at,
    get_tei_xpath_text_content_list,
    iter_layout_block_tei_children,
    tei_xpath
)


class TeiAuthor(TeiElementWrapper):
    pass


class TeiAffiliation(TeiElementWrapper):
    pass


class TeiSectionParagraph(TeiElementWrapper):
    def __init__(self, element: etree.ElementBase):
        super().__init__(element)
        self._pending_whitespace: Optional[str] = None

    def add_content(self, layout_block: LayoutBlock):
        if self._pending_whitespace:
            extend_element(self.element, [self._pending_whitespace])
        self._pending_whitespace = layout_block.whitespace
        extend_element(
            self.element,
            iter_layout_block_tei_children(layout_block)
        )


class TeiSection(TeiElementWrapper):
    def get_title_text(self) -> str:
        return '\n'.join(get_tei_xpath_text_content_list(
            self.element,
            '//tei:head',
        ))

    def get_paragraph_text_list(self) -> List[str]:
        return get_tei_xpath_text_content_list(
            self.element,
            '//tei:p',
        )

    def add_paragraph(self, paragraph: TeiSectionParagraph):
        self.element.append(paragraph.element)

    def create_paragraph(self) -> TeiSectionParagraph:
        return TeiSectionParagraph(TEI_E('p'))


class TeiDocument(TeiElementWrapper):
    def __init__(self, root: Optional[etree.ElementBase] = None):
        if root is None:
            self.root = TEI_E('TEI')
        else:
            self.root = root
        self._reference_element: Optional[etree.ElementBase] = None
        super().__init__(self.root)

    def get_or_create_element_at(self, path: List[str]) -> etree.ElementBase:
        return get_or_create_element_at(self.root, path)

    def set_child_element_at(self, path: List[str], child: etree.ElementBase):
        parent = self.get_or_create_element_at(path)
        parent.append(child)

    def get_title(self) -> str:
        return '\n'.join(get_tei_xpath_text_content_list(
            self.root,
            '//tei:fileDesc/tei:titleStmt/tei:title[@level="a"][@type="main"]',
        ))

    def set_title(self, title: str):
        self.set_child_element_at(
            ['teiHeader', 'fileDesc', 'titleStmt'],
            TEI_E('title', title, level="a", type="main")
        )

    def set_title_layout_block(self, title_block: LayoutBlock):
        self.set_child_element_at(
            ['teiHeader', 'fileDesc', 'titleStmt'],
            TEI_E(
                'title',
                {'level': 'a', 'type': 'main'},
                *iter_layout_block_tei_children(title_block)
            )
        )

    def get_abstract(self) -> str:
        return '\n'.join(get_tei_xpath_text_content_list(
            self.root,
            '//tei:abstract/tei:p',
        ))

    def set_abstract(self, abstract: str):
        self.set_child_element_at(
            ['teiHeader', 'profileDesc', 'abstract'],
            TEI_E('p', abstract)
        )

    def set_abstract_layout_block(self, abstract_block: LayoutBlock):
        self.set_child_element_at(
            ['teiHeader', 'profileDesc', 'abstract'],
            TEI_E('p', *iter_layout_block_tei_children(abstract_block))
        )

    def get_body_element(self) -> etree.ElementBase:
        return self.get_or_create_element_at(['text', 'body'])

    def get_body(self) -> TeiElementWrapper:
        return TeiElementWrapper(self.get_body_element())

    def get_back_element(self) -> etree.ElementBase:
        return self.get_or_create_element_at(['text', 'back'])

    def get_back_annex_element(self) -> etree.ElementBase:
        return self.get_or_create_element_at(['text', 'back', 'div[@type="annex"]'])

    def get_back_annex(self) -> TeiElementWrapper:
        return TeiElementWrapper(self.get_back_annex_element())

    def get_references_element(self) -> etree.ElementBase:
        if self._reference_element is not None:
            return self._reference_element
        self._reference_element = self.get_or_create_element_at(
            ['text', 'back', 'div[@type="references"]']
        )
        return self._reference_element

    def get_references(self) -> TeiElementWrapper:
        return TeiElementWrapper(self.get_references_element())

    def get_body_sections(self) -> List[TeiSection]:
        return [
            TeiSection(element)
            for element in tei_xpath(self.get_body_element(), './tei:div')
        ]

    def add_body_section(self, section: TeiSection):
        self.get_body_element().append(section.element)

    def add_back_annex_section(self, section: TeiSection):
        self.get_back_annex_element().append(section.element)

    def add_acknowledgement_section(self, section: TeiSection):
        section.element.attrib['type'] = 'acknowledgement'
        self.get_back_element().append(section.element)

    def create_section(self) -> TeiSection:
        return TeiSection(TEI_E('div'))



import logging
from sciencebeam_parser.models.citation.training_data import (
    CitationTeiTrainingDataGenerator,
    CitationTrainingTeiParser
)

from sciencebeam_parser.models.data import (
    DocumentFeaturesContext
)
from sciencebeam_parser.models.model import Model
from sciencebeam_parser.models.citation.data import CitationDataGenerator
from sciencebeam_parser.models.citation.extract import CitationSemanticExtractor


LOGGER = logging.getLogger(__name__)


class CitationModel(Model):
    def get_data_generator(
        self,
        document_features_context: DocumentFeaturesContext
    ) -> CitationDataGenerator:
        return CitationDataGenerator(
            document_features_context=document_features_context
        )

    def get_semantic_extractor(self) -> CitationSemanticExtractor:
        return CitationSemanticExtractor()

    def get_tei_training_data_generator(self) -> CitationTeiTrainingDataGenerator:
        return CitationTeiTrainingDataGenerator()

    def get_training_tei_parser(self) -> CitationTrainingTeiParser:
        return CitationTrainingTeiParser()



import os
import logging
from typing import Sequence

from sciencebeam_trainer_delft.utils.download_manager import DownloadManager
from sciencebeam_trainer_delft.utils.io import is_external_location


LOGGER = logging.getLogger(__name__)


def download_if_url_from_alternatives(
    download_manager: DownloadManager,
    alternative_file_url_or_path_list: Sequence[str]
) -> str:
    for file_url_or_path in alternative_file_url_or_path_list:
        if not is_external_location(file_url_or_path):
            if os.path.exists(file_url_or_path):
                return file_url_or_path
            LOGGER.debug('local file doesnt exist: %r', file_url_or_path)
            continue
        local_file = download_manager.get_local_file(file_url_or_path)
        if os.path.exists(local_file):
            return local_file
    LOGGER.debug(
        'no existing local files found, downloading: %r', alternative_file_url_or_path_list
    )
    for file_url_or_path in alternative_file_url_or_path_list:
        try:
            local_file = download_manager.download_if_url(file_url_or_path)
            if os.path.exists(local_file):
                return local_file
            LOGGER.debug(
                'local file for %r not found: %r',
                file_url_or_path, local_file
            )
        except FileNotFoundError:
            LOGGER.debug('remote file not found: %r', file_url_or_path)
    raise FileNotFoundError('no file found for %r' % alternative_file_url_or_path_list)



import os
import codecs
from contextlib import contextmanager
from typing import Iterable, Sequence
from urllib.parse import urlparse

import fsspec

from sciencebeam_trainer_delft.utils.io import (
    auto_uploading_output_file as _auto_uploading_output_file,
    is_external_location,
    open_file
)


def get_file_system_protocol_for_url(url: str) -> fsspec.AbstractFileSystem:
    parsed_url = urlparse(url)
    return parsed_url.scheme or 'file'


def get_file_system_for_url(url: str) -> fsspec.AbstractFileSystem:
    return fsspec.filesystem(get_file_system_protocol_for_url(url))


def get_file_system_protocols(
    fs: fsspec.AbstractFileSystem
) -> Sequence[str]:
    return (fs.protocol,) if isinstance(fs.protocol, str) else fs.protocol


def get_file_system_default_protocol(
    fs: fsspec.AbstractFileSystem
) -> str:
    return get_file_system_protocols(fs)[0]


def get_fully_qualified_path_for_protocol_and_path(
    protocol: str,
    path: str
) -> str:
    if 'file' in protocol:
        return path
    return f'{protocol}://{path}'


def iter_fully_qualified_paths_for_protocol_and_paths(
    protocol: str,
    paths: Iterable[str]
) -> Iterable[str]:
    return (
        get_fully_qualified_path_for_protocol_and_path(protocol, path)
        for path in paths
    )


def get_fully_qualified_path_for_fs_and_path(
    fs: fsspec.AbstractFileSystem,
    path: str
) -> str:
    return get_fully_qualified_path_for_protocol_and_path(
        get_file_system_default_protocol(fs),
        path
    )


def glob(
    glob_pattern: str
) -> Sequence[str]:
    protocol = get_file_system_protocol_for_url(glob_pattern)
    fs: fsspec.AbstractFileSystem = fsspec.filesystem(protocol)
    return list(iter_fully_qualified_paths_for_protocol_and_paths(
        protocol,
        fs.glob(glob_pattern)
    ))


def makedirs(
    path: str,
    exist_ok: bool = False
):
    get_file_system_for_url(path).makedirs(path, exist_ok=exist_ok)


@contextmanager
def auto_uploading_binary_output_file(filepath: str, **kwargs):
    if not is_external_location(filepath):
        # Note: the upstream implementation doesn't currently auto-compress local files
        file_dirname = os.path.dirname(filepath)
        if file_dirname:
            os.makedirs(os.path.dirname(filepath), exist_ok=True)
        with open_file(filepath, mode='wb', **kwargs) as fp:
            yield fp
            return
    with _auto_uploading_output_file(filepath, 'wb', **kwargs) as fp:
        yield fp


@contextmanager
def auto_uploading_text_output_file(filepath: str, encoding: str, **kwargs):
    with auto_uploading_binary_output_file(filepath, **kwargs) as fp:
        yield codecs.getwriter(encoding)(fp)


def auto_uploading_output_file(filepath: str, mode: str, encoding: str = 'utf-8', **kwargs):
    if mode == 'w':
        return auto_uploading_text_output_file(filepath, encoding=encoding, **kwargs)
    if mode == 'wb':
        return auto_uploading_binary_output_file(filepath, **kwargs)
    raise ValueError('invalid mode: %r' % mode)


def write_bytes(filepath: str, data: bytes, **kwargs):
    with auto_uploading_output_file(filepath, mode='wb', **kwargs) as fp:
        fp.write(data)


def write_text(filepath: str, text: str, encoding: str, **kwargs):
    # Note: the upstream implementation doesn't support encoding with compression
    write_bytes(
        filepath,
        codecs.encode(text, encoding=encoding),
        **kwargs
    )



import logging
import os
import socket
import time
from contextlib import closing
from threading import Lock, current_thread
from typing import Optional, Sequence

from sciencebeam_parser.utils.background_process import (
    BackgroundProcess,
    ChildProcessReturnCodeError,
    CommandRestartableBackgroundProcess,
    exec_with_logging
)

from .office_scripts import get_office_script_directory
from .office_scripts.office_utils import find_pyuno_office, get_start_listener_command


LOGGER = logging.getLogger(__name__)


def change_ext(path: str, old_ext: Optional[str], new_ext: str) -> str:
    if old_ext is None:
        old_ext = os.path.splitext(path)[1]
        if old_ext == '.gz':
            path = path[:-len(old_ext)]
            old_ext = os.path.splitext(path)[1]
    if old_ext and path.endswith(old_ext):
        return path[:-len(old_ext)] + new_ext
    return path + new_ext


class UnoConnectionError(ConnectionError):
    pass


def _exec_pyuno_script(
    script_filename: str,
    args: Sequence[str],
    process_timeout: Optional[float] = None,
    daemon: bool = False
) -> BackgroundProcess:
    if not os.path.exists(script_filename):
        from glob import glob  # pylint: disable=import-outside-toplevel
        LOGGER.info(
            '%s does not exist, found: %s',
            script_filename,
            list(glob('%s/**/*' % os.path.dirname(script_filename)))
        )
        raise RuntimeError('%s does not exist' % script_filename)
    office = find_pyuno_office()
    command = [
        office.uno_python_path,
        script_filename
    ] + list(args)
    env = {'PYTHONPATH': office.uno_path}
    LOGGER.info('executing: %s (env: %s)', command, env)
    try:
        p = exec_with_logging(
            command,
            env=env,
            logging_prefix='converter',
            process_timeout=process_timeout,
            daemon=daemon
        )
    except ChildProcessReturnCodeError as e:
        if e.returncode == 9:
            raise UnoConnectionError('failed to connect to uno server: %s' % e.returncode) from e
        raise type(e)(
            'failed to run converter: %s' % e.returncode,
            returncode=e.returncode
        ) from e
    return p


def _exec_doc_converter(
    args: Sequence[str],
    enable_debug: bool = False,
    process_timeout: Optional[float] = None,
    daemon: bool = False
) -> BackgroundProcess:
    office_scripts_directory = get_office_script_directory()
    doc_converter_script_filename = os.path.abspath(os.path.join(
        office_scripts_directory,
        'doc_converter.py'
    ))
    if enable_debug:
        args = ['--debug'] + list(args)
    return _exec_pyuno_script(
        doc_converter_script_filename,
        args,
        process_timeout=process_timeout,
        daemon=daemon
    )


class ListenerProcess(CommandRestartableBackgroundProcess):
    def __init__(self, port: int, host: str = '127.0.0.1', connect_timeout: int = 10):
        super().__init__(
            command=get_start_listener_command(port=port),
            name='listener on port %s' % port,
            logging_prefix='listener[port:%s]' % port,
            stop_at_exit=True
        )
        self.port = port
        self.host = host
        self.connect_timeout = connect_timeout

    def __repr__(self) -> str:
        return (
            '{type_name}('
            'port={self.port}'
            ', host={self.host}'
            ', connect_timeout={self.connect_timeout}'
            ', command={command}'
            ', process={self.process}'
            ')'
        ).format(
            type_name=type(self).__name__,
            self=self,
            command=repr(self.command)
        )

    def is_alive(self) -> bool:
        if not self.is_running():
            return False
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
            sock.settimeout(self.connect_timeout)  # pylint: disable=no-member
            if sock.connect_ex((self.host, self.port)) == 0:  # pylint: disable=no-member
                return True
        return False

    def wait_for_is_alive(self, timeout: float) -> bool:
        start_time = time.monotonic()
        while not self.is_alive():
            if not self.is_running():
                return False
            if time.monotonic() - start_time >= timeout:
                return False
            time.sleep(0.5)
        return True

    def start_and_check_alive(self, timeout=10, **kwargs) -> None:
        super().start(**kwargs)
        if self.wait_for_is_alive(timeout=timeout):
            return
        self.stop_if_running()
        assert self.process is not None
        if self.process.returncode == 81:
            # see https://bugs.documentfoundation.org/show_bug.cgi?id=107912
            #     "headless firstrun crashes (exit code 81)"
            LOGGER.info('detected first-run error code 81, re-trying..')
            self.start_and_check_alive(timeout=timeout, **kwargs)
            return
        raise ConnectionError('failed to start listener (unable to connect)')

    def start_listener_if_not_running(self, max_uptime: float = None, **kwargs) -> None:
        if self.is_alive():
            uptime = self.get_uptime()
            if not max_uptime or uptime <= max_uptime:
                return
            LOGGER.info('stopping listener, exceeded max uptime: %.3f > %.3f', uptime, max_uptime)
            self.stop()
        self.start_and_check_alive(**kwargs)


class DocConverterWrapper:  # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        port: int = 2003,
        enable_debug: bool = False,
        no_launch: bool = True,
        keep_listener_running: bool = True,
        process_timeout: Optional[float] = None,
        max_uptime: float = 10,
        stop_listener_on_error: bool = True
    ):
        self.port = port
        self.enable_debug = enable_debug
        self.no_launch = no_launch
        self.keep_listener_running = keep_listener_running
        self.process_timeout = process_timeout
        self.max_uptime = max_uptime
        self.stop_listener_on_error = stop_listener_on_error
        self._listener_process = ListenerProcess(port=port)
        self._lock = Lock()
        self._concurrent_count = 0

    def __repr__(self) -> str:
        return (
            '{type_name}('
            'port={self.port}'
            ', keep_listener_running={self.keep_listener_running}'
            ', _listener_process={self._listener_process}'
            ')'
        ).format(
            type_name=type(self).__name__,
            self=self
        )

    def start_listener_if_not_running(self) -> None:
        self._listener_process.start_listener_if_not_running(max_uptime=self.max_uptime)

    def stop_listener_if_running(self) -> None:
        self._listener_process.stop_if_running()

    def _do_convert(
        self,
        temp_source_filename: str,
        output_type: str = 'pdf',
        remove_line_no: bool = True,
        remove_header_footer: bool = True,
        remove_redline: bool = True
    ) -> str:
        if self.no_launch:
            self.start_listener_if_not_running()

        temp_target_filename = change_ext(
            temp_source_filename, None, '-output.%s' % output_type
        )

        args = []
        args.extend([
            'convert',
            '--format', output_type
        ])
        if remove_line_no:
            args.append('--remove-line-no')
        if remove_header_footer:
            args.append('--remove-header-footer')
        if remove_redline:
            args.append('--remove-redline')
        args.extend([
            '--port', str(self.port),
            '--output-file', str(temp_target_filename),
            temp_source_filename
        ])
        if self.no_launch:
            args.append('--no-launch')
        if self.keep_listener_running:
            args.append('--keep-listener-running')
        try:
            _exec_doc_converter(
                args,
                enable_debug=self.enable_debug,
                process_timeout=self.process_timeout
            )
        except UnoConnectionError:
            self.stop_listener_if_running()
            raise
        except Exception:
            if self.stop_listener_on_error:
                self.stop_listener_if_running()
            raise

        if not os.path.exists(temp_target_filename):
            raise RuntimeError('temp target file missing: %s' % temp_target_filename)
        return temp_target_filename

    def convert(self, *args, **kwargs) -> str:
        thread_id = current_thread().ident
        try:
            self._concurrent_count += 1
            LOGGER.debug(
                'attempting to aquire lock, thread id: %s, concurrent count: %s',
                thread_id, self._concurrent_count
            )
            with self._lock:
                LOGGER.debug(
                    'aquired lock, thread id: %s, concurrent count: %s',
                    thread_id, self._concurrent_count
                )
                return self._do_convert(*args, **kwargs)
        finally:
            self._concurrent_count -= 1
            LOGGER.debug(
                'exiting convert (released lock if it was aquired),'
                ' thread id: %s, concurrent count: %s',
                thread_id, self._concurrent_count
            )



import os
import logging
import threading
from typing import Iterable, Optional, List, Tuple

import numpy as np

from sciencebeam_trainer_delft.sequence_labelling.engines.wapiti import WapitiWrapper
from sciencebeam_trainer_delft.utils.io import copy_file
from sciencebeam_trainer_delft.utils.download_manager import DownloadManager
from sciencebeam_trainer_delft.sequence_labelling.engines.wapiti_adapters import (
    WapitiModelAdapter,
    WapitiModel
)

from sciencebeam_parser.app.context import AppContext
from sciencebeam_parser.models.model_impl import ModelImpl
from sciencebeam_parser.utils.download import download_if_url_from_alternatives
from sciencebeam_parser.utils.lazy import LazyLoaded


LOGGER = logging.getLogger(__name__)


class WapitiServiceModelAdapter(WapitiModelAdapter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._lock = threading.Lock()
        self._wapiti_timeout = 20.0
        self._wapiti_timeout_counter = 0
        self._wapiti_trial_count = 10

    @staticmethod
    def load_from(
            model_path: str,
            download_manager: DownloadManager,
            wapiti_binary_path: str = None) -> 'WapitiModelAdapter':
        # overriding method to return WapitiServiceModelAdapter
        model_file_path = os.path.join(model_path, 'model.wapiti.gz')
        model_file_paths = [model_file_path, os.path.splitext(model_file_path)[0]]
        LOGGER.debug('checking for existing local model files: %r', model_file_paths)
        local_model_file_path = download_if_url_from_alternatives(
            download_manager=download_manager,
            alternative_file_url_or_path_list=model_file_paths
        )
        LOGGER.debug('local_model_file_path: %s', local_model_file_path)
        if local_model_file_path.endswith('.gz'):
            local_uncompressed_file_path = os.path.splitext(local_model_file_path)[0]
            copy_file(local_model_file_path, local_uncompressed_file_path, overwrite=False)
            local_model_file_path = local_uncompressed_file_path
        return WapitiServiceModelAdapter(
            WapitiWrapper(
                wapiti_binary_path=wapiti_binary_path
            ),
            model_file_path=local_model_file_path,
            model_path=model_path
        )

    def stop(self):
        wapiti_model = self._wapiti_model
        if wapiti_model is None:
            return
        self._wapiti_model = None
        LOGGER.info('stopping wapiti process: %s', wapiti_model.process.pid)
        wapiti_model.process.kill()

    def on_wapiti_timeout(self):
        self._wapiti_timeout_counter += 1
        LOGGER.info(
            'wapiti timeout (%s, counter=%d)',
            self._wapiti_timeout, self._wapiti_timeout_counter
        )
        self.stop()

    def _get_tag_results_with_timeout(
        self,
        x: np.ndarray,
        features: np.ndarray,
        output_format: str = None
    ) -> List[List[Tuple[str, str]]]:
        prev_wapiti_timeout_counter = self._wapiti_timeout_counter
        timer = threading.Timer(self._wapiti_timeout, self.on_wapiti_timeout)
        timer.start()
        result = list(self.iter_tag_using_model(x, features, output_format))
        timer.cancel()
        if self._wapiti_timeout_counter != prev_wapiti_timeout_counter:
            raise TimeoutError('wapiti timeout received during processing')
        return result

    def _get_tag_results_with_timeout_and_retry(
        self,
        x: np.ndarray,
        features: np.ndarray,
        output_format: str = None
    ) -> List[List[Tuple[str, str]]]:
        attempt = 0
        while True:
            try:
                return self._get_tag_results_with_timeout(x, features, output_format)
            except Exception as exc:  # pylint: disable=broad-except
                attempt += 1
                LOGGER.warning(
                    'received error processing data: %r, attempt=%d/%d, texts=%r',
                    exc, attempt, self._wapiti_trial_count, list(x), exc_info=True
                )
                if attempt >= self._wapiti_trial_count:
                    LOGGER.warning('final attempt, re-raising exception')
                    raise

    def iter_tag(
        self,
        x: np.ndarray,
        features: np.ndarray,
        output_format: str = None
    ) -> Iterable[List[Tuple[str, str]]]:
        # by default, WapitiModelAdapter will run the binary for each call
        # using "iter_tag_using_model" will result in a wapiti process
        # that we communicate with via stdin / stdout
        with self._lock:
            yield from self._get_tag_results_with_timeout_and_retry(x, features, output_format)


class WapitiModelImpl(ModelImpl):
    def __init__(self, model_url: str, app_context: AppContext):
        self.model_url = model_url
        self.app_context = app_context
        self._lazy_model = LazyLoaded[WapitiModelAdapter](self._load_model)

    def __repr__(self) -> str:
        return '%s(%r, loaded=%r)' % (
            type(self).__name__, self.model_url, self._lazy_model.is_loaded
        )

    def _load_model(self) -> WapitiModel:
        model = WapitiServiceModelAdapter.load_from(
            self.model_url,
            wapiti_binary_path=self.app_context.lazy_wapiti_binary_wrapper.get_binary_path(),
            download_manager=self.app_context.download_manager
        )
        LOGGER.info('loaded wapiti model: %r', self.model_url)
        return model

    @property
    def model(self) -> WapitiModel:
        return self._lazy_model.get()

    def preload(self):
        self._lazy_model.get()

    def predict_labels(
        self,
        texts: List[List[str]],
        features: List[List[List[str]]],
        output_format: Optional[str] = None
    ) -> List[List[Tuple[str, str]]]:
        model = self.model
        result = model.tag(texts, features=features, output_format=output_format)
        token_count = sum(len(text) for text in texts)
        LOGGER.info(
            'predicted labels using wapiti model (document count: %d, token count: %d)',
            len(texts), token_count
        )
        return result
